"""
How does this work?
When you fetch the page of a user on instagram you get an html with javascript embedded
in it with a json object embedded in that. This json object describes the user, his id,
his profile photo and the first 12 images for that user.
If you want more you have to do a follow-up AJAX request to the server.
"""
import json
import logging
import time

import pyeventroute.route
from lxml import etree

import pyscrapers.core.utils
from pyscrapers.core.url_set import UrlSet


def is_rate_limit(response) -> bool:
    """
    Rate limit messages look like this:
    { "message": "rate limited", "status": "fail" }
    :param response:
    :return:
    """
    return response["status"] == "fail" and response["message"] == "rate limited"


def scrape_instagram(user_id: str, session, url_set: UrlSet) -> None:
    domain = 'www.instagram.com'
    base = 'https://{domain}'.format(domain=domain)
    url = '{base}/{user_id}/'.format(base=base, user_id=user_id)

    logger = logging.getLogger(__name__)

    response = session.get(url)
    root = pyscrapers.core.utils.get_html_dom_content(response)
    # scrape.utils.print_element(root)

    # register regular expressions with lxml
    # this means that we can use regular expression functions like 'match'
    # by specifying 're:match' in our xpath expressions
    ns = etree.FunctionNamespace("http://exslt.org/regular-expressions")
    ns.prefix = 're'
    e_a = root.xpath('//script[re:match(text(), "^window._sharedData")]')
    assert len(e_a) == 1
    data = e_a[0].text
    json_text = data[data.find('{'):data.rfind('}') + 1]
    d = json.loads(json_text)
    my_list = d['entry_data']['ProfilePage']
    assert len(my_list) == 1
    c = my_list[0]["graphql"]["user"]
    if 'profile_pic_url_hd' in c:
        url_set.append(c['profile_pic_url_hd'])
    elif 'profile_pic_url' in c:
        url_set.append(c['profile_pic_url'])
    user_id = c['id']
    url2 = '{base}/graphql/query/'.format(base=base)
    count = 50
    query_hashes = [
        'bd0d6d184eefd4d0ce7036c11ae58ed9',  # posts
        'ff260833edf142911047af6024eb634a',  # tagged
    ]
    keys = [
        'edge_owner_to_timeline_media',
        'edge_user_to_photos_of_you',
    ]
    stats_video = 0
    stats_image = 0
    stats_shortcode_video = 0
    for query_hash, key in zip(query_hashes, keys):
        logger.debug("size of list is [{}]".format(len(url_set.urls_list)))
        has_next_page = True
        end_cursor = None
        while has_next_page:
            variables = {
                    'id': user_id,
                    'first': count,
            }
            if end_cursor:
                variables['after'] = end_cursor
            params = {
                'query_hash': query_hash,
                'variables': json.dumps(variables)
            }
            response = session.get(url2, params=params).json()
            data_user = response['data']['user'][key]
            has_next_page = data_user['page_info']['has_next_page']
            end_cursor = data_user['page_info']['end_cursor']
            for outer_node in data_user['edges']:
                inner_node = outer_node['node']
                if inner_node['is_video'] and 'video_url' in inner_node:
                    url_set.append(inner_node['video_url'])
                    stats_video += 1
                if inner_node['is_video'] and 'shortcode' in inner_node:
                    params = {
                        'query_hash': '03f541f086ce0a9b31f67688ff9c1e09',
                        'shortcode': inner_node['shortcode'],
                    }
                    response_short = session.get(url2, params=params).json()
                    while is_rate_limit(response_short):
                        time.sleep(60)
                        response_short = session.get(url2, params=params).json()
                    if 'data' in response_short:
                        url_set.append(response_short['data']['shortcode_media']['video_url'])
                    else:
                        pyeventroute.route.accept(response_short)
                    stats_shortcode_video += 1
                if 'display_url' in inner_node:
                    url_set.append(inner_node['display_url'])
                    stats_image += 1
    logger.info("stats_video [{}]".format(stats_video))
    logger.info("stats_image [{}]".format(stats_image))
    logger.info("stats_shortcode_video [{}]".format(stats_shortcode_video))



"""
Module to handle scraping of pornhub.

References:
- https://pypi.org/project/pornhub-api/
"""
import logging
from itertools import islice
from typing import List

import pornhub_api
import requests
from pornhub_api import PornhubApi

import pyscrapers.core.utils
from pyscrapers.configs import ConfigPornhubSearch, ConfigDebugRequests, ConfigUrl, get_cookies
from pyscrapers.core.url_set import UrlSet
from pyscrapers.workers.youtube_dl_handlers import youtube_dl_download_url, youtube_dl_download_urls


def print_stars_all(api: pornhub_api.api.PornhubApi) -> None:
    stars_results = api.stars.all()
    for star in islice(stars_results.stars, 0, ConfigPornhubSearch.limit):
        print(star.star.star_name)


def print_stars_all_detailed(api: pornhub_api.api.PornhubApi) -> None:
    stars_results = api.stars.all_detailed()
    for star in islice(stars_results.stars, 0, ConfigPornhubSearch.limit):
        print(star.star.star_name)


def print_categories(api: pornhub_api.api.PornhubApi) -> None:
    categories = api.video.categories()
    for category in categories.categories:
        print("id [{}], name [{}]".format(category.id, category.category))


def print_tags(api: pornhub_api.api.PornhubApi) -> None:
    tags = api.video.tags(ConfigPornhubSearch.literal)
    for tag in tags:
        print("tag [{}]".format(tag))


def download_search() -> None:
    """
    Download movies from pornhub
    """
    logger = logging.getLogger(__name__)
    api = PornhubApi()

    limit = ConfigPornhubSearch.limit
    page = 1
    counter = 0
    errors = 0
    exceptions = []
    except_urls = []
    while True:
        kwargs = dict()
        if ConfigPornhubSearch.use_ordering:
            kwargs["ordering"] = ConfigPornhubSearch.ordering
        if ConfigPornhubSearch.use_period:
            kwargs["period"] = ConfigPornhubSearch.period
        if ConfigPornhubSearch.use_tags:
            kwargs["tags"] = ConfigPornhubSearch.tags
        try:
            data = api.search.search(
                ConfigPornhubSearch.query,
                page=page,
                **kwargs,
            )
        except ValueError as e:
            code = e.args[0]["code"]
            if code == "2001":  # no videos found (end of results)
                break
            raise e
        urls = [video.url for video in data.videos]
        if limit is not None:
            urls = list(islice(urls, 0, limit - counter))
        for url in urls:
            logger.info("doing item [{}]".format(counter))
            # noinspection PyBroadException
            try:
                youtube_dl_download_url(url)
            except Exception as e:
                errors += 1
                exceptions.append(e)
                except_urls.append(url)
            counter += 1
        page += 1
        if counter == limit:
            break
    if errors > 0:
        logger.info("number of errors [{}]".format(errors))
        logger.info("except_urls [{}]".format(except_urls))


def get_number_of_pages(root) -> int:
    """
    return number of pages for a pornstar
    :param root:
    :return:
    """
    counters = root.xpath('//div[contains(@class,\'showingInfo\')]')
    assert len(counters) == 1
    counter = counters[0]
    number = int(counter.text.strip().split()[3])
    number_in_page = 36
    number_of_pages = -(-number // number_in_page)
    return number_of_pages


def get_urls_from_page(root) -> List[str]:
    """
    return urls from page
    :param root:
    :return:
    """
    # video_sections = root.xpath('//section[@id=\'videosTab\']')
    # assert len(video_sections) == 1, len(video_sections)
    # video_section = video_sections[0]
    # video_section = video_sections[0]
    # <ul class="dropdownHottestVideos videos" id="hottestMenuSection">
    # <ul class="videos row-5-thumbs" id="mostRecentVideosSection">
    # pyscrapers.core.utils.print_element(root)
    # import sys
    # sys.exit(1)
    xpath_attempts = [
        '//ul[contains(@class,\'pornstarsVideos\')]',
        '//ul[@id=\'moreData\']',
        '//ul[@id=\'mostRecentVideosSection\']',
    ]
    video_sections = []
    for xpath_attempt in xpath_attempts:
        video_sections = root.xpath(xpath_attempt)
        if len(video_sections) == 1:
            break
    if len(video_sections) == 1:
        video_section = video_sections[0]
        elements = video_section.xpath('li[contains(@class,\'pcVideoListItem\')]')
        urls = []
        for element in elements:
            # pyscrapers.core.utils.print_element(element)
            # key = element.attrib['_vkey']
            key = element.attrib['data-video-vkey']
            url = "https://www.pornhub.com/view_video.php?viewkey={key}".format(key=key)
            urls.append(url)
        return urls
    return []


def url_generator(url: str):
    yield url
    page = 2
    while True:
        yield "{url}?page={page}".format(url=url, page=page)
        page += 1


def download_url() -> None:
    if ConfigDebugRequests.debug:
        pyscrapers.core.utils.debug_requests()
    session = requests.Session()
    session.cookies = get_cookies()
    logger = logging.getLogger(__name__)
    urls = UrlSet()
    for url in url_generator(url=ConfigUrl.url):
        logger.info("getting [{}]...".format(url))
        page = session.get(url)
        if page.status_code != 200:
            break
        root = pyscrapers.core.utils.get_html_dom_content(page)
        new_urls = get_urls_from_page(root)
        if len(new_urls) == 0:
            break
        logger.info("got [{}] new urls".format(len(new_urls)))
        for i, new_url in enumerate(new_urls):
            logger.info("url {} is [{}]".format(i, new_url))
        urls.extend(new_urls)
    session.close()
    logger.info("got total [{}] urls".format(len(urls.urls_list)))
    logger.info("got [{}] appended twice urls".format(urls.appended_twice))
    youtube_dl_download_urls(urls.urls_list)



"""
All configurations for pyscrapers
"""
import os
import logging

import browser_cookie3
from pytconf import Config, ParamCreator


class ConfigLogging(Config):
    """
    Parameters to control logging
    """
    loglevel = ParamCreator.create_choice(
        choice_list=[
            logging.getLevelName(logging.NOTSET),
            logging.getLevelName(logging.DEBUG),
            logging.getLevelName(logging.INFO),
            logging.getLevelName(logging.WARNING),
            logging.getLevelName(logging.WARN),
            logging.getLevelName(logging.ERROR),
            logging.getLevelName(logging.FATAL),
            logging.getLevelName(logging.CRITICAL),
        ],
        help_string="What log level to use?",
        default=logging.getLevelName(logging.ERROR),

    )


class ConfigDebugRequests(Config):
    """
    Configure how to user the requests module
    """
    debug = ParamCreator.create_bool(
        help_string="Do you want to debug the requests module?",
        default=False,
    )


class ConfigDownload(Config):
    """
    Configure details about the download process
    """
    download_as_collecting = ParamCreator.create_bool(
        help_string="Do you want download while collecting the urls?",
        default=False,
    )
    download = ParamCreator.create_bool(
        help_string="really download or just print the urls?",
        default=True,
    )
    folder = ParamCreator.create_existing_folder(
        help_string="where to save the data to?",
        default=".",
    )


class ConfigCookiesSource(Config):
    """
        Configure where to get cookies from
    """

    browser = ParamCreator.create_choice(
        choice_list=["none", "firefox", "chrome"],
        help_string="Which browser to take cookies from?",
        default="firefox",
    )


def get_cookies():
    if ConfigCookiesSource.browser == "firefox":
        return browser_cookie3.firefox()
    if ConfigCookiesSource.browser == "chrome":
        return browser_cookie3.chrome()
    raise ValueError(f"unsupported browser [{ConfigCookiesSource.browser}")


class ConfigSiteId(Config):
    """
    Parameters for downloading workers
    """
    site = ParamCreator.create_choice(
        choice_list=["facebook", "instagram", "travelgirls", "vk", "mamba.ru"],
        help_string="Which site to download from?",
    )
    user_id = ParamCreator.create_str(
        help_string="""Which user id to user?
            https://www.facebook.com/profile.php?id=[user_id]
            https://www.instagram.com/[user_id]
            http://www.travelgirls.com/member/[user_id]
            https://vk.com/id[user_id]
            http://www.mamba.ru/mb[user_id]""",
    )


class ConfigPornhubSearch(Config):
    """
    Parameters for search
    """
    query = ParamCreator.create_str(
        help_string="What is the query string?",
    )
    use_ordering = ParamCreator.create_bool(
        help_string="use ordering in the search operation",
        default=True,
    )
    ordering = ParamCreator.create_choice(
        choice_list=["longest", "featured", "newest", "mostviewed", "rating"],
        help_string="by which ordering to fetch result?",
        default="longest",
    )
    use_period = ParamCreator.create_bool(
        help_string="use period in the search operation",
        default=False,
    )
    period = ParamCreator.create_choice(
        choice_list=["weekly", "monthly", "alltime"],
        help_string="what period to search?",
        default="weekly",
    )
    use_tags = ParamCreator.create_bool(
        help_string="should we use tags in search?",
        default=False,
    )
    tags = ParamCreator.create_list_str(
        help_string="tags to be used in search",
        default=[],
    )
    literal = ParamCreator.create_str(
        help_string="literal for tags (one character)",
        default="f",
    )
    limit = ParamCreator.create_int_or_none(
        help_string="Limit on search results or None for no limit",
        default=100,
    )


class ConfigYoutubeDl(Config):
    """
    Configuration for youtube downloads
    """
    use_archive = ParamCreator.create_bool(
        help_string="Should we use an archive?",
        default=True,
    )
    archive_file = ParamCreator.create_existing_file(
        help_string="What file to use as archive?",
        default=os.path.expanduser('~/.config/youtube-dl-archive'),
    )


class ConfigUrl(Config):
    """
    Parameters for what url to download
    """
    url = ParamCreator.create_str(
        help_string="url to download (e.g. https://www.pornhub.com/model/lily)"
    )



import pyscrapers.core.utils
from pyscrapers.core.url_set import UrlSet


def scrape_travelgirls(user_id: str, session, url_set: UrlSet) -> None:
    main_url = 'https://www.travelgirls.com/member/{user_id}'.format(user_id=user_id)
    r = session.get(main_url)
    root = pyscrapers.core.utils.get_html_dom_content(r)

    e_a = root.xpath('//a[contains(@class,\'photo\')]')
    for x in e_a:
        # print(etree.tostring(x, pretty_print=True))
        children = x.getchildren()
        assert len(children) == 1
        img = children[0]
        url = pyscrapers.core.utils.add_http(img.attrib['src'], main_url)
        url = url.replace('mini', '')
        url = url.replace('thumb', '')
        url_set.append(url)



"""
This is a module that returns information about video streams based on running ffprobe
as a subprocess.

References:
- http://stackoverflow.com/questions/3844430/how-to-get-the-duration-of-a-video-in-python/3844467
"""

import json
import subprocess


def probe(vid_file_path):
    """ Give a json from ffprobe command line

    @vid_file_path : The absolute (full) path of the video file, string.
    """
    if isinstance(vid_file_path, str):
        raise Exception('Give ffprobe a full file path of the video')

    args = [
        "ffprobe",
        "-loglevel",
        "quiet",
        "-print_format",
        "json",
        # "-show_format",
        "-show_streams",
        "-select_streams",
        "v",
        vid_file_path,
    ]
    return json.loads(subprocess.check_output(args).decode())


def width(vid):
    _json = probe(vid)
    return int(_json["streams"][0]["width"])


def height(vid):
    _json = probe(vid)
    return int(_json["streams"][0]["height"])


def duration(vid_file_path):
    """ Video's duration in seconds, return a float number
    """
    _json = probe(vid_file_path)

    if 'format' in _json:
        if 'duration' in _json['format']:
            return float(_json['format']['duration'])

    if 'streams' in _json:
        # commonly stream 0 is the video
        for s in _json['streams']:
            if 'duration' in s:
                return float(s['duration'])

    # if everything didn't happen,
    # we got here because no single 'return' in the above happen.
    raise Exception('I found no duration')



import logging
import os
import shutil
import urllib.parse
from typing import Union, List

from pyscrapers.configs import ConfigDownload


class UrlSet:
    """ set of urls, with no duplicates. Can be downloaded """
    def __init__(self):
        """
        constructor
        """
        self.urls_set = set()
        self.urls_list = []
        self.appended_twice = 0
        self.counter_jpg = 0
        self.counter_mp4 = 0

    def append(self, url: str) -> None:
        """
        add url to the list
        :param url:
        :return:
        """
        logger = logging.getLogger(__name__)
        logger.debug("collecting [{}]".format(url))
        if url in self.urls_set:
            self.appended_twice += 1
        else:
            self.urls_set.add(url)
            self.urls_list.append(url)

    def extend(self, urls: List[str]) -> None:
        for url in urls:
            self.append(url)

    def print(self) -> None:
        """
        print the list
        :return:
        """
        for url in self.urls_list:
            print(url)

    def suggest_filename(self, suffix: str) -> Union[str, None]:
        if suffix == ".jpg":
            filename = 'image{0:04}.jpg'.format(self.counter_jpg)
            self.counter_jpg += 1
            return filename
        if suffix == ".mp4":
            filename = 'video{0:04}.mp4'.format(self.counter_mp4)
            self.counter_mp4 += 1
            return filename
        logger = logging.getLogger(__name__)
        logger.error('do not know how to handle suffix [%s]...', suffix)
        return None

    def get_filename(self, suffix: str) -> str:
        filename = self.suggest_filename(suffix)
        while os.path.isfile(filename):
            filename = self.suggest_filename(suffix)
        return filename

    def download(self, session):
        """
        download the list
        :param session:
        :return:
        """
        logger = logging.getLogger(__name__)
        logger.info('got [%d] urls', len(self.urls_list))
        if ConfigDownload.download:
            for url in self.urls_list:
                parse_result = urllib.parse.urlparse(url)
                path = parse_result.path
                logger.info('downloading [%s]...', url)
                response = session.get(url, stream=True)
                assert response.status_code == 200, response.content

                filename = self.get_filename(os.path.splitext(path)[1])
                if filename is None:
                    continue

                with open(filename, 'wb') as file_handle:
                    response.raw.decode_content = True
                    shutil.copyfileobj(response.raw, file_handle)
                logger.info('written [%s]...', filename)



"""
Module that handles the interaction with the youtube_dl library

References:
- https://github.com/ytdl-org/youtube-dl/blob/master/README.md#embedding-youtube-dl
- https://github.com/ytdl-org/youtube-dl/blob/3e4cedf9e8cd3157df2457df7274d0c842421945/youtube_dl/YoutubeDL.py#L137-L312
"""
from typing import List

import youtube_dl

from pyscrapers.configs import ConfigYoutubeDl, ConfigUrl


class MyLogger:
    def debug(self, msg):
        pass

    def warning(self, msg):
        pass

    # noinspection PyMethodMayBeStatic
    def error(self, msg):
        print(msg)


def youtube_dl_handler() -> None:
    youtube_dl_download_url(ConfigUrl.url)


def youtube_dl_download_url(url: str) -> None:
    youtube_dl_download_urls([url])


def youtube_dl_download_urls(urls: List[str]) -> None:
    ydl_opts = {
        'format': 'bestaudio/best',
        # 'logger': MyLogger(),
        'nooverwrites': True,
        'ignoreerrors': True,
    }
    if ConfigYoutubeDl.use_archive:
        ydl_opts['download_archive'] = ConfigYoutubeDl.archive_file
    with youtube_dl.YoutubeDL(ydl_opts) as ydl:
        ydl.download(urls)



import json
import logging
import sys

import lxml.etree
import lxml.html

from pyscrapers.core.url_set import UrlSet


def get_my_content(r):
    """
    the return from the server in vk is not a standard HTML.
    this is why we must cut it up and cant use the regular
    'get_real_content' helper.
    """
    assert r.status_code == 200
    # str_content=r.content.decode(errors='ignore')
    try:
        content: bytes = r.content
        str_content = content.decode(errors='ignore')
    except Exception as e:
        print(e)
        print('could not decode')
        print(r.content)
        sys.exit(1)
    str_content = str_content[str_content.find('<input'):]
    c = str.encode('<html><body>') + str.encode(str_content) + str.encode('</body></html>')
    root = lxml.html.fromstring(c)
    return root


def scrape_vk(user_id: str, session, url_set: UrlSet) -> None:
    logger = logging.getLogger(__name__)
    url = 'https://vk.com/al_photos.php'
    data = {
        'act': 'show_albums',
        'al': '1',
        'owner': user_id,
    }
    r = session.post(url, data=data)
    root = get_my_content(r)

    e_albums = root.xpath('//div[@class="photos_album_title_wrap"]')
    albums = dict()
    for x in e_albums:
        name_of_album = x.getparent().getparent().getparent().getparent().attrib['id'].split('?')[0]
        e_len = x.xpath('.//div[@class="photos_album_counter fl_r"]/text()')
        albums[name_of_album] = int(e_len[0])
    logger.debug(albums)
    total_images = 0
    for v in albums.values():
        logger.debug('got [%d] partial images', v)
        total_images += v
    logger.debug('got [%d] potential images', total_images)

    count = 0
    got = 1
    while got:
        got = 0
        data = {
            'act': 'show_albums',
            'al': '2',
            'owner': user_id,
            'offset': count,
        }
        logger.debug('doing request %d', count)
        r = session.post(url, data=data)
        root = get_my_content(r)
        e_a = root.xpath('//a[@onclick]')
        for x in e_a:
            onclick = x.attrib['onclick']
            if onclick.startswith('return showPhoto'):
                json_str = onclick[onclick.find('{'):onclick.rfind('}') + 1]
                # bas string, need fix lots of things...
                json_str = json_str.replace('\'', '"')
                json_str = json_str.replace('jumpTo', '"jumpTo"')
                json_str = json_str.replace('z:', '"z":')
                json_obj = json.loads(json_str)
                base = json_obj['temp']['base']
                if base == '':
                    continue
                largest = 0
                largest_url = None
                min_len = min(len(v) for k, v in json_obj['temp'].items() if k.endswith('_'))
                if min_len == 3:
                    for k, v in json_obj['temp'].items():
                        if k != 'base':
                            size = v[1] * v[2]
                            if size > largest:
                                largest = size
                                largest_url = v[0]
                    full_url = base + largest_url + '.jpg'
                    url_set.append(full_url)
                    got += 1
                if min_len == 1:
                    for k, v in json_obj['temp'].items():
                        if k != 'base':
                            add_url = v[0]
                            full_url = base + add_url + '.jpg'
                            url_set.append(full_url)
                    got += 1
        count += got



"""
The default group of operations that pyscrapers has
"""
import logging
import shelve

import pylogconf.core
import requests
from pornhub_api import PornhubApi
from pytconf import register_endpoint, register_main, config_arg_parse_and_launch

import pyscrapers.core.utils
from pyscrapers.configs import ConfigDebugRequests, ConfigCookiesSource, ConfigSiteId, ConfigPornhubSearch, \
    ConfigYoutubeDl, ConfigDownload, ConfigLogging, ConfigUrl, get_cookies
from pyscrapers.core.url_set import UrlSet
from pyscrapers.static import APP_NAME, VERSION_STR
from pyscrapers.workers.drumeo import get_number_of_pages, get_courses, get_course_details, get_course_urls, \
    download_course
from pyscrapers.workers.facebook import scrape_facebook
from pyscrapers.workers.instagram import scrape_instagram
from pyscrapers.workers.mambaru import scrape_mambaru
from pyscrapers.workers.pornhub import download_search, print_stars_all_detailed, download_url
from pyscrapers.workers.travelgirls import scrape_travelgirls
from pyscrapers.workers.vk import scrape_vk
from pyscrapers.workers.youtube_dl_handlers import youtube_dl_handler


@register_endpoint(
    description="Download photo albums from various sites",
    configs=[
        ConfigSiteId,
        ConfigDebugRequests,
        ConfigCookiesSource,
        ConfigDownload,
        ConfigLogging,
    ],
)
def photos():
    if ConfigDebugRequests.debug:
        pyscrapers.core.utils.debug_requests()
    logger = logging.getLogger('pyscrapers')
    logger.setLevel(ConfigLogging.loglevel)
    session = requests.Session()
    session.cookies = get_cookies()
    url_set = UrlSet()
    if ConfigSiteId.site == "facebook":
        scrape_facebook(ConfigSiteId.user_id, session, url_set)
    if ConfigSiteId.site == "instagram":
        scrape_instagram(ConfigSiteId.user_id, session, url_set)
    if ConfigSiteId.site == "travelgirls":
        scrape_travelgirls(ConfigSiteId.user_id, session, url_set)
    if ConfigSiteId.site == "vk":
        scrape_vk(ConfigSiteId.user_id, session, url_set)
    if ConfigSiteId.site == "mamba.ru":
        scrape_mambaru(ConfigSiteId.user_id, session, url_set)
    url_set.download(session)
    session.close()


@register_endpoint(
    description="Download videos from drumeo",
    configs=[
        ConfigDebugRequests,
        ConfigCookiesSource,
    ],
)
def drumeo():
    if ConfigDebugRequests.debug:
        pyscrapers.core.utils.debug_requests()

    session = requests.Session()
    session.cookies = get_cookies()

    logger = logging.getLogger(__name__)
    courses = False
    reload = {}
    with shelve.open("cache.db") as d:
        if "courses" in d:
            list_of_courses = d["courses"]
            print("got from cache [{}] courses".format(len(list_of_courses)))
        else:
            pages = get_number_of_pages(courses=courses, session=session)
            print("number of pages is [{}]".format(pages))
            list_of_courses = get_courses(pages, courses=courses, session=session)
            print("got [{}] courses".format(len(list_of_courses)))
            d["courses"] = list_of_courses
        for i, course in enumerate(list_of_courses):
            logger.info("course number [%s]", i)
            if course.number in d and course.number not in reload:
                list_of_courses[i] = d[course.number]
                logger.info("got from cache [%s]", list_of_courses[i])
            else:
                get_course_details(course, courses=courses, session=session)
                get_course_urls(course, courses=courses, session=session)
                print(course)
                d[course.number] = course
            download_course(list_of_courses[i], session)
    session.close()


@register_endpoint(
    description="print stars all detailed",
)
def pornhub_stars_all_detailed():
    api = PornhubApi()
    print_stars_all_detailed(api)


@register_endpoint(
    description="Download search results from pornhub",
    configs=[
        ConfigPornhubSearch,
        ConfigDownload,
        ConfigYoutubeDl,
    ],
)
def pornhub_download_search():
    download_search()


@register_endpoint(
    description="Download url videos from pornhub",
    configs=[
        ConfigDebugRequests,
        ConfigCookiesSource,
        ConfigUrl,
        ConfigDownload,
        ConfigYoutubeDl,
    ],
)
def pornhub_download_url():
    download_url()


@register_endpoint(
    description="Download movies using youtuble_dl",
    configs=[
        ConfigUrl,
        ConfigYoutubeDl,
    ],
)
def youtube_dl():
    youtube_dl_handler()


@register_main(
    main_description="pyscapers will help you download stuff from the web",
    app_name=APP_NAME,
    version=VERSION_STR,
)
def main():
    pylogconf.core.setup()
    config_arg_parse_and_launch()


if __name__ == '__main__':
    main()



"""
download photos from facebook
"""


import logging

import pyscrapers.core.utils
from pyscrapers.core.url_set import UrlSet


def scrape_facebook(user_id: str, session, url_set: UrlSet) -> None:
    """
    download photos from facebook
    :param user_id:
    :param session:
    :param url_set:
    :return:
    """
    logger = logging.getLogger(__name__)

    url = 'https://www.facebook.com/{user_id}/workers'.format(user_id=user_id)
    logger.debug('url is [%s]', url)
    result = session.get(url)
    root = pyscrapers.core.utils.get_html_dom_content(result)

    elements_img = root.xpath('//img')
    url_set.append(elements_img)



"""
This module is s set of utilities for this entire project
"""


import http.client
import logging
import os
import shutil
import urllib.parse
from typing import List

import requests
import lxml
import lxml.html

import pyscrapers.core.ffprobe


def print_cookies(cookies, domain):
    """
    Print the cookies from a specific domain
    :param cookies:
    :param domain:
    :return:
    """
    print(domain)
    for cookie in cookies:
        print(cookie)


def get_http_status_string(code: int):
    """
    This function returns a description of an HTTP status code (404 - not found etc).
    Unfortunately, the requests module does not provide a clean API for this so
    we must access a protected member (underscore member) of 'requests.status_code'.
    See:
    https://stackoverflow.com/questions/24718557/get-the-description-of-a-status-code-in-python-requests
    :param code:
    :return:
    """
    return "http code [{}], [{}]".format(code, requests.status_codes.codes[code][0])


def get_html_dom_content(response):
    """
    Get the content from a request
    :param response:
    :return:
    """
    assert response.status_code == 200, get_http_status_string(response.status_code)
    str_content = response.content.decode()
    root = lxml.html.fromstring(str_content)
    return root


def download_urls(session, urls: List[str], start=0):
    """
    Download a list of urls
    :param session:
    :param urls:
    :param start:
    :return:
    """
    logger = logging.getLogger(__name__)
    counter = start
    logger.info('got [%d] real urls', len(urls))
    for url in urls:
        parse_result = urllib.parse.urlparse(url)
        path = parse_result.path
        logger.info('downloading [%s]...', url)
        response = session.get(url, stream=True)
        assert response.status_code == 200, response.content

        filename = None
        if path.endswith(".jpg"):
            filename = 'image{0:04}.jpg'.format(counter)
        if path.endswith(".mp4"):
            filename = 'video{0:04}.mp4'.format(counter)
        if filename is None:
            logger.error('do not know how to handle path [%s]...', path)
            continue
        assert not os.path.isfile(filename), "already have filename {}".format(path)
        with open(filename, 'wb') as file_handle:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file_handle)
        logger.info('written [%s]...', filename)
        counter += 1


def debug_requests():
    """
    Activate the debugging features of the requests module
    :return:
    """
    http.client.HTTPConnection.debuglevel = 1
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.propagate = True


def add_http(url, main_url):
    """
    add two urls together
    :param url:
    :param main_url:
    :return:
    """
    return urllib.parse.urljoin(main_url, url)


def print_element(element):
    """
    from xml elements from etree
    :param element:
    :return:
    """
    print(lxml.etree.tostring(element, pretty_print=True).decode())


def download_url(session, source: str, target: str) -> None:
    """
    Download a single url to a file
    """
    logger = logging.getLogger(__name__)
    logger.info('downloading [%s] to [%s]', source, target)
    if os.path.isfile(target):
        logger.info('skipping [%s]', target)
        return
    try:
        response = session.get(source, stream=True)
        assert response.status_code == 200, response.content
        with open(target, 'wb') as file_handle:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file_handle)
    except IOError:
        os.unlink(target)
    logger.info('written [%s]...', target)


FAIL = True


def download_video_if_wider(session, source: str, target: str, width: int) -> bool:
    """
    Download a video if it is wider than a certain width
    :param source:
    :param target:
    :param width:
    :param session:
    :return:
    """
    logger = logging.getLogger(__name__)
    logger.info('downloading [%s] to [%s]', source, target)
    if os.path.isfile(target):
        file_width = pyscrapers.core.ffprobe.height(target)
        if file_width >= width:
            logger.info('skipping because video with width exists [%s] %s %s', target, file_width, width)
            return True
        logger.info('continuing with download because of width [%s] %s %s', target, file_width, width)
    try:
        response = session.get(source, stream=True)
        if FAIL:
            assert response.status_code == 200, response.content
        else:
            if response.status_code != 200:
                logger.info("got bad error code [%s] and failed to download", response.status_code)
                return False
        with open(target, 'wb') as file_handle:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file_handle)
    except IOError as e:
        if os.path.isfile(target):
            os.unlink(target)
        if FAIL:
            raise ValueError("count not download") from e
        logger.info("failed to download file")
        return False
    logger.info('written [%s]...', target)
    return True
